// file      : xsd/cxx/tree/name-processor.cxx
// author    : Boris Kolpackov <boris@codesynthesis.com>
// copyright : Copyright (c) 2006-2007 Code Synthesis Tools CC
// license   : GNU GPL v2 + exceptions; see accompanying LICENSE file

#include <cxx/tree/name-processor.hxx>
#include <cxx/tree/elements.hxx>

#include <cult/containers/set.hxx>
#include <cult/containers/map.hxx>

#include <iostream>

using std::endl;

namespace CXX
{
  namespace Tree
  {
    namespace
    {
      //
      //
      typedef Cult::Containers::Set<String> NameSet;

      class Context: public Tree::Context
      {
      public:
        Context (CLI::Options const& options,
                 Counts const& counts,
                 Boolean generate_xml_schema,
                 SemanticGraph::Schema& root,
                 SemanticGraph::Path const& file)
            : Tree::Context (std::wcerr,
                             root,
                             options,
                             counts,
                             generate_xml_schema),
              schema_path_ (file),
              schema_path (schema_path_),
              ns_ (0),
              ns (ns_),
              global_type_names (global_type_names_),
              global_element_names (global_element_names_)
        {
        }

      protected:
        Context (Context& c)
            : Tree::Context (c),
              schema_path (c.schema_path),
              ns (c.ns),
              global_type_names (c.global_type_names),
              global_element_names (c.global_element_names)
        {
        }

      private:
        SemanticGraph::Path const schema_path_;
        SemanticGraph::Namespace* ns_;

        Cult::Containers::Map<String, NameSet> global_type_names_;
        Cult::Containers::Map<String, NameSet> global_element_names_;

      public:
        SemanticGraph::Path const& schema_path;
        SemanticGraph::Namespace*& ns;

        Cult::Containers::Map<String, NameSet>& global_type_names;
        Cult::Containers::Map<String, NameSet>& global_element_names;
      };

      String
      find_name (SemanticGraph::Nameable& n, NameSet& set)
      {
        String base_name (Tree::Context::escape (n.name ()));
        String name (base_name);

        for (UnsignedLong i (1); set.find (name) != set.end (); ++i)
        {
          std::wostringstream os;
          os << i;
          name = base_name + os.str ();
        }

        set.insert (name);
        n.context ().set ("name", name);

        return name;
      }


      //
      //
      struct Enumerator: Traversal::Enumerator
      {
        Enumerator (NameSet& set)
            : set_ (set)
        {
        }

        virtual Void
        traverse (Type& e)
        {
          find_name (e, set_);
        }

      private:
        NameSet& set_;
      };


      //
      //
      struct Enumeration: Traversal::Enumeration
      {
        Enumeration (String const& name_ = "")
            : name (name_)
        {
        }

        virtual Void
        traverse (Type& e)
        {
          if (e.named ())
            name = e.name ();

          NameSet enumerator_set;
          enumerator_set.insert (name);

          Enumerator enumerator (enumerator_set);
          Traversal::Names names (enumerator);

          Traversal::Enumeration::names (e, names);
        }

      private:
        String name;
      };

      //
      //
      template <typename Complex>
      struct Member: Traversal::Member
      {
        Member (NameSet& set)
            : set_ (set)
        {
        }

        virtual Void
        traverse (Type& m)
        {
          if (Tree::Context::skip (m))
            return;

          String name (find_name (m, set_));

          // Handle anonymous types.
          //
          SemanticGraph::Type& t (m.type ());

          if (!t.named () && !t.context ().count ("seen"))
          {
            Complex complex (name);
            Enumeration enumeration (name);
            Traversal::Belongs belongs;

            belongs >> complex;
            belongs >> enumeration;

            t.context ().set ("seen", true);
            Traversal::Member::belongs (m, belongs);
            t.context ().remove ("seen");
          }
        }

      private:
        NameSet& set_;
      };


      //
      //
      struct Complex: Traversal::Complex
      {
        Complex (String const& name_ = "")
            : name (name_)
        {
        }

        virtual Void
        traverse (Type& c)
        {
          if (c.named ())
            name = c.name ();

          NameSet member_set;
          member_set.insert (name);

          // Add our base's members to the initial list.
          //
          if (c.inherits_p ())
          {
            // @@ What if this types name is the same as one of base's
            //    members?
            //
            SemanticGraph::Type& base (c.inherits ().base ());

            if (base.is_a<SemanticGraph::Complex> ())
            {
              if (!base.context ().count (
                    "cxx-tree-name-processor-member-set"))
              {
                Complex t; // Base type is always named.
                t.dispatch (base);
              }

              NameSet const& base_set (
                base.context ().get<NameSet> (
                  "cxx-tree-name-processor-member-set"));

              member_set.insert (base_set.begin (), base_set.end ());
            }
          }

          Member<Complex> member (member_set);
          Traversal::Names names (member);

          Traversal::Complex::names (c, names);

          // @@ I am not removing this at the moment.
          //
          c.context ().set ("cxx-tree-name-processor-member-set",
                            member_set);
        }

      private:
        String name;
      };


      //
      //
      struct GlobalType: Traversal::Type
      {
        GlobalType (NameSet& set)
            : set_ (set)
        {
        }

        virtual Void
        traverse (SemanticGraph::Type& t)
        {
          find_name (t, set_);
        }

      private:
        NameSet& set_;
      };


      //
      //
      struct GlobalElement: Traversal::Element,
                            GlobalElementBase,
                            Context
      {
        GlobalElement (Context& c,
                       NameSet const& type_set,
                       NameSet& element_set)
            : GlobalElementBase (c),
              Context (c),
              type_set_ (type_set),
              element_set_ (element_set)
        {
        }

        virtual Void
        traverse (Type& e)
        {
          // First we need to figure out if we need to process this
          // global element.
          //
          if (!generate_p (e))
            return;

          // Make sure the name is unique among global elements and
          // does not collide with a global type name.
          //
          String base (Tree::Context::escape (e.name ()));
          String name (base);

          // If we are conflicting with a type name let's first try to
          // simply append an underscore and only resort to ugly names
          // like name1, etc. if this fails.
          //
          if (type_set_.find (name) != type_set_.end ())
            name += L"_";

          for (UnsignedLong i (1);
               element_set_.find (name) != element_set_.end () ||
                 type_set_.find (name) != type_set_.end ();
               ++i)
          {
            std::wostringstream os;
            os << i;
            name = base + os.str ();
          }

          //if (name != base)
          //  os << "renamed '" << base << "' to '" << name << "'" << endl;

          element_set_.insert (name);
          e.context ().set ("name", name);


          // Handle anonymous type.
          //
          SemanticGraph::Type& t (e.type ());

          if (!t.named ())
          {
            Complex complex (name);
            Enumeration enumeration (name);
            Traversal::Belongs belongs;

            belongs >> complex;
            belongs >> enumeration;

            t.context ().set ("seen", true);
            Traversal::Element::belongs (e, belongs);
            t.context ().remove ("seen");
          }
        }

        /*
      private:
        SemanticGraph::Type*
        find (SemanticGraph::Schema& schema, String const& name)
        {
          using SemanticGraph::Type;
          using SemanticGraph::Scope;
          using SemanticGraph::Namespace;

          String ns_name (ns->name ());

          // Get all namespaces across include/import hierarchy with
          // our namespace name.
          //
          // @@ using root schema here.
          //
          Scope::NamesIteratorPair nip (schema.find (ns_name));

          for (; nip.first != nip.second; ++nip.first)
          {
            Namespace& ns (dynamic_cast<Namespace&> (nip.first->named ()));

            Scope::NamesIteratorPair types (ns.find (name));

            for (; types.first != types.second; ++types.first)
            {
              if (Type* t = dynamic_cast<Type*> (&types.first->named ()))
              {
                return t;
              }
            }
          }

          return 0;
        }
        */

      private:
        NameSet const& type_set_;
        NameSet& element_set_;
      };

      struct NamespacePassOne: Traversal::Namespace, Context
      {
        NamespacePassOne (Context& c)
            : Context (c)
        {
        }

        virtual Void
        traverse (Type& ns)
        {
          NameSet& type_set (global_type_names[ns.name ()]);

          GlobalType type (type_set);
          Traversal::Names names (type);

          Context::ns = &ns;

          Traversal::Namespace::names (ns, names);
          Traversal::Namespace::names (ns);

          Context::ns = 0;
        }
      };


      struct NamespacePassTwo: Traversal::Namespace, Context
      {
        NamespacePassTwo (Context& c)
            : Context (c)
        {
        }

        virtual Void
        traverse (Type& ns)
        {
          String const& name (ns.name ());

          NameSet const& type_set (global_type_names[name]);
          NameSet& element_set (global_element_names[name]);

          GlobalElement element (*this, type_set, element_set);
          Traversal::Names names (element);

          Context::ns = &ns;

          Traversal::Namespace::names (ns, names);

          Context::ns = 0;
        }
      };


      // Go into implied/included/imported schemas while making sure
      // we don't process the same stuff more than once.
      //
      struct UsesPassOne: Traversal::Uses
      {
        virtual Void
        traverse (Type& u)
        {
          SemanticGraph::Schema& s (u.schema ());

          if (!s.context ().count ("cxx-tree-name-processor-seen-pass-1"))
          {
            s.context ().set ("cxx-tree-name-processor-seen-pass-1", true);
            Traversal::Uses::traverse (u);
          }
        }
      };

      struct UsesPassTwo: Traversal::Uses
      {
        virtual Void
        traverse (Type& u)
        {
          SemanticGraph::Schema& s (u.schema ());

          if (!s.context ().count ("cxx-tree-name-processor-seen-pass-2"))
          {
            s.context ().set ("cxx-tree-name-processor-seen-pass-2", true);
            Traversal::Uses::traverse (u);
          }
        }
      };

      Void
      process_impl (CLI::Options const& ops,
                    SemanticGraph::Schema& tu,
                    SemanticGraph::Path const& file)
      {
        Counts counts;
        Context ctx (ops, counts, false, tu, file);

        // Pass one - handle types.
        //
        {
          Traversal::Schema schema;
          UsesPassOne uses;

          schema >> uses >> schema;

          Traversal::Names schema_names;
          NamespacePassOne ns (ctx);
          Traversal::Names ns_names;

          schema >> schema_names >> ns >> ns_names;

          Complex complex;
          Enumeration enumeration;

          ns_names >> complex;
          ns_names >> enumeration;

          // Some twisted schemas do recusive inclusions.
          //
          tu.context ().set ("cxx-tree-name-processor-seen-pass-1", true);

          schema.dispatch (tu);
        }

        // Pass two - handle global elements (and their anonymous types).
        //
        {
          Traversal::Schema schema;
          UsesPassTwo uses;

          schema >> uses >> schema;

          Traversal::Names schema_names;
          NamespacePassTwo ns (ctx);
          Traversal::Names ns_names;

          schema >> schema_names >> ns;

          // Some twisted schemas do recusive inclusions.
          //
          tu.context ().set ("cxx-tree-name-processor-seen-pass-2", true);

          schema.dispatch (tu);
        }
      }
    }

    Void NameProcessor::
    process (CLI::Options const& ops,
             SemanticGraph::Schema& tu,
             SemanticGraph::Path const& file)
    {
      process_impl (ops, tu, file);
    }
  }
}
