// file      : xsd/cxx/tree/tree-inline.cxx
// author    : Boris Kolpackov <boris@codesynthesis.com>
// copyright : Copyright (c) 2005-2007 Code Synthesis Tools CC
// license   : GNU GPL v2 + exceptions; see accompanying LICENSE file

#include <cxx/tree/tree-inline.hxx>

#include <xsd-frontend/semantic-graph.hxx>
#include <xsd-frontend/traversal.hxx>

#include <backend-elements/regex.hxx>

namespace CXX
{
  namespace Tree
  {
    namespace
    {
      struct List : Traversal::List, protected virtual Context
      {
        List (Context& c, String const& name_ = L"")
            : Context (c), name (name_)
        {
        }

        virtual Void
        traverse (Type& l)
        {
          if (l.named ())
          {
            name = ename (l);

            // If renamed name is empty then we do not need to generate
            // anything for this type.
            //
            if (renamed_type (l, name) && !name)
              return;
          }


          enter_scope (name);
          String scoped_name (scope);
          leave_scope ();

          String base (L"::xsd::cxx::tree::list< " +
                       item_type_name (l) +
                       L", " + char_type + L" >");

          os << "// " << scoped_name << endl
             << "//" << endl
             << endl;

          // c-tor ()
          //
          os << inl
             << scoped_name << "::" << endl
             << name << " ()"
             << "{"
             << "}";

          // copy c-tor ()
          //
          os << inl
             << scoped_name << "::" << endl
             << name << " (const " << name << "& o," << endl
             << flags_type << " f," << endl
             << any_type << "* c)" << endl
             << ": " << any_simple_type << " (o, f, c)," << endl
             << base << " (o, f, c)"
             << "{"
             << "}";
        }

      private:
        String
        item_type_name (Type& l)
        {
          // This type is always named.
          //
          std::wostringstream o;

          MemberTypeName type (*this, o);
          type.dispatch (l.argumented ().type ());

          return o.str ();
        }

      private:
        String name;
      };


      struct Union : Traversal::Union, protected virtual Context
      {
        Union (Context& c, String const& name_ = L"")
            : Context (c), name (name_)
        {
        }

        virtual Void
        traverse (Type& u)
        {
          if (u.named ())
          {
            name = ename (u);

            // If renamed name is empty then we do not need to generate
            // anything for this type.
            //
            if (renamed_type (u, name) && !name)
              return;
          }


          String base (xs_ns_name () + L"::string");

          enter_scope (name);
          String scoped_name (scope);
          leave_scope ();

          os << "// " << scoped_name << endl
             << "//" << endl
             << endl;

          if (options.value<CLI::generate_default_ctor> ())
          {
            // c-tor ()
            //
            os << inl
               << scoped_name << "::" << endl
               << name << " ()" << endl
               << ": " << base << " ()"
               << "{"
               << "}";
          }

          // c-tor (string const&)
          //
          os << inl
             << scoped_name << "::" << endl
             << name << " (const " << string_type << "& s)" << endl
             << ": " << base << " (s)"
             << "{"
             << "}";

          // copy c-tor ()
          //
          os << inl
             << scoped_name << "::" << endl
             << name << " (const " << name << "& o," << endl
             << flags_type << " f," << endl
             << any_type << "* c)" << endl
             << ": " << base << " (o, f, c)"
             << "{"
             << "}";
        }

      private:
        String name;
      };

      // Enumeration
      //

      // Generate a sequence of explicit c-tor calls until we reach
      // one of the fundamental string types that can be constructed
      // from char literals.
      //
      struct CtorCallSequence: Traversal::Complex,
                               Traversal::Fundamental::Type,
                               protected virtual Context
      {
        CtorCallSequence (Context& c, String const& arg)
            : Context (c), arg_ (arg), base_type_name_ (c)
        {
        }

        virtual Void
        traverse (SemanticGraph::Complex& c)
        {
          // This type should be ultimately string based.
          //
          assert (c.inherits_p ());

          os << ename (c) << " (" << endl;

          dispatch (c.inherits ().base ());

          os << ")";
        }

        virtual Void
        traverse (SemanticGraph::Fundamental::Type& t)
        {
          base_type_name_.dispatch (t);

          os << " (" << arg_ << ")";
        }

      private:
        String arg_;
        BaseTypeName base_type_name_;
      };

      struct Enumeration : Traversal::Enumeration,
                           protected virtual Context
      {
        Enumeration (Context& c, String const& name_ = L"")
            : Context (c), name (name_), member_ (c)
        {
          inherits_member_ >> member_;
        }

        virtual Void
        traverse (Type& e)
        {
          if (e.named ())
          {
            name = ename (e);

            // If renamed name is empty then we do not need to generate
            // anything for this type.
            //
            if (renamed_type (e, name) && !name)
              return;
          }


          Boolean string_based (false);
          {
            IsStringBasedType t (string_based);
            t.dispatch (e);
          }

          Boolean enum_based (false);
          if (string_based)
          {
            SemanticGraph::Enumeration* be (0);
            IsEnumBasedType t (be);
            t.dispatch (e);

            enum_based = (be != 0);
          }

          String base; // base type name
          {
            std::wostringstream o;

            BaseTypeName base_type (*this, o);
            Traversal::Inherits inherits_type (base_type);

            inherits (e, inherits_type);
            base = o.str ();
          }

          enter_scope (name);
          String scoped_name (scope);
          leave_scope ();

          os << "// " << scoped_name << endl
             << "// " << endl
             << endl;

          // default c-tor
          //
          if (options.value<CLI::generate_default_ctor> ())
          {
            os << inl
               << scoped_name << "::" <<  endl
               << name << " ()" << endl
               << ": " << base << " ()"
               << "{"
               << "}";
          }

          // c-tor (_xsd_name)
          //
          if (string_based)
          {
            os << inl
               << scoped_name << "::" <<  endl
               << name << " (_xsd_" << name << " v)" << endl
               << ": ";

            // If we are enum-based then we can just use the corresponding
            // base c-tor directly. Otherwise we will use the from-string
            // c-tor.
            //
            if (enum_based)
              os << base << " (v)";
            else
            {
              CtorCallSequence t (*this, L"_xsd_" + name + L"_literals_[v]");
              t.dispatch (e.inherits ().base ());
            }

            os << "{"
               << "}";
          }


          // c-tor (base)
          //
          os << inl
             << scoped_name << "::" <<  endl
             << name << " (const ";

          inherits (e, inherits_member_);

          os << "& _xsd_" << name << ")" << endl
             << ": " << base << " (_xsd_" << name << ")"
             << "{"
             << "}";


          // copy c-tor
          //
          os << inl
             << scoped_name << "::" <<  endl
             << name << " (const " << name << "& v," << endl
             << flags_type << " f," << endl
             << any_type << "* c)" << endl
             << ": " << base << " (v, f, c)"
             << "{"
             << "}";

          // operato= (_xsd_name)
          //
          if (string_based)
          {
            os << inl
               << scoped_name << "& " << scoped_name << "::" << endl
               << "operator= (_xsd_" << name << " v)"
               << "{"
               << "static_cast< " << base << "& > (*this) = ";

            // If we are enum-based then we can just use the corresponding
            // base assignment directly. Otherwise we will use the from-
            // base assignment and a from-string base c-tor.
            //
            if (enum_based)
              os << "v";
            else
            {
              os << endl;

              CtorCallSequence t (*this, L"_xsd_" + name + L"_literals_[v]");
              t.dispatch (e.inherits ().base ());
            }

            os << ";"
               << endl
               << "return *this;"
               << "}";
          }

          os << endl;
        }

      private:
        String name;

        Traversal::Inherits inherits_member_;
        MemberTypeName member_;
      };

      struct MemberFunction: Traversal::Member, protected virtual Context
      {
        MemberFunction (Context& c)
            : Context (c)
        {
        }

        virtual Void
        traverse (Type& m)
        {
          if (skip (m)) return;

          String const& name (ename (m));

          Boolean fund (false);
          {
            IsFundamentalType t (fund);
            t.dispatch (m.type ());
          }

          if (max (m) != 1)
          {
            // sequence
            //
            String container (iscope_name (name, "container"));
            String q_container (scope + L"::" + container);

            // container const&
            // name () const;
            //
            os << inl
               << "const " << q_container << "& " << scope << "::" << endl
               << name << " () const"
               << "{"
               << "return this->_xsd_" << name << "_;"
               << "}";

            // container&
            // name ();
            //
            os << inl
               << q_container << "& " << scope << "::" << endl
               << name << " ()"
               << "{"
               << "return this->_xsd_" << name << "_;"
               << "}";

            // void
            // name (container const&);
            //
            os << inl
               << "void " << scope << "::" << endl
               << name << " (const " << container << "& " << name << ")"
               << "{"
               << "this->_xsd_" << name << "_ = " << name << ";"
               << "}";
          }
          else if (min (m) == 0 &&
                   (!m.is_a<SemanticGraph::Attribute> () || !m.default_ ()))
          {
            // optional
            //
            String type (iscope_name (name, "type"));
            String container (iscope_name (name, "container"));
            String q_container (scope + L"::" + container);

            // container const&
            // name () const;
            //
            os << inl
               << "const " << q_container << "& " << scope << "::" << endl
               << name << " () const"
               << "{"
               << "return this->_xsd_" << name << "_;"
               << "}";

            // container&
            // name ();
            //
            os << inl
               << q_container << "& " << scope << "::" << endl
               << name << " ()"
               << "{"
               << "return this->_xsd_" << name << "_;"
               << "}";

            // void
            // name (type const&);
            //
            os << inl
               << "void " << scope << "::" << endl
               << name << " (const " << type << "& " << name << ")"
               << "{"
               << "this->_xsd_" << name << "_.set (" << name << ");"
               << "}";

            // void
            // name (container const&);
            //
            os << inl
               << "void " << scope << "::" << endl
               << name << " (const " << container << "& " << name << ")"
               << "{"
               << "this->_xsd_" << name << "_ = " << name << ";"
               << "}";

            // void
            // name (auto_ptr<type>);
            //
            if (!fund)
              os << inl
                 << "void " << scope << "::" << endl
                 << name << " (::std::auto_ptr< " << type << " > " <<
                name << ")"
                 << "{"
                 << "this->_xsd_" << name << "_.set (" << name << ");"
                 << "}";
          }
          else
          {
            // one
            //
            String type (iscope_name (name, "type"));
            String q_type (scope + L"::" + type);

            // type const&
            // name () const;
            //
            os << inl
               << "const " << q_type << "& " << scope << "::" << endl
               << name << " () const"
               << "{"
               << "return this->_xsd_" << name << "_.get ();"
               << "}";

            // type&
            // name ();
            //
            os << inl
               << q_type << "& " << scope << "::" << endl
               << name << " ()"
               << "{"
               << "return this->_xsd_" << name << "_.get ();"
               << "}";

            // void
            // name (type const&);
            //
            os << inl
               << "void " << scope << "::" << endl
               << name << " (const " << type << "& " << name << ")"
               << "{"
               << "this->_xsd_" << name << "_.set (" << name << ");"
               << "}";

            // void
            // name (auto_ptr<type>);
            //
            if (!fund)
              os << inl
                 << "void " << scope << "::" << endl
                 << name << " (::std::auto_ptr< " << type << " > " <<
                name << ")"
                 << "{"
                 << "this->_xsd_" << name << "_.set (" << name << ");"
                 << "}";
          }
        }
      };

      template <typename Complex>
      struct Member : Traversal::Member,
                      protected virtual Context
      {
        Member (Context& c)
            : Context (c),
              member_function_ (c)
        {
        }

        virtual Void
        traverse (Type& m)
        {
	  if (skip (m)) return;

          String const& name (ename (m));
          SemanticGraph::Type& t (m.type ());

          // Go after anonymous type first.
          //
          if (!t.named () && !t.context ().count ("seen"))
          {
            Traversal::Belongs belongs_;

            List list (*this, name);
            Union union_ (*this, name);
            Complex complex (*this, name);
            Enumeration enumeration (*this, name);

            belongs_ >> list;
            belongs_ >> union_;
            belongs_ >> complex;
            belongs_ >> enumeration;

            enter_scope (name);
            enter_scope (L"_xsd_" + name + L"_");

            t.context ().set ("seen", true);

            belongs (m, belongs_);

            t.context ().remove ("seen");

            leave_scope ();
            leave_scope ();
          }

          // default_value
          //
          if (m.default_ ())
          {
            Boolean simple (true);

            if (m.is_a<SemanticGraph::Element> ())
            {
              IsSimpleType test (simple);
              test.dispatch (t);
            }

            if (simple)
            {
              String type (iscope_name (name, "type"));
              String q_type (scope + L"::" + type);

              os << inl
                 << "const " << q_type << "& " << scope << "::" <<
                name << "::" << endl
                 << unclash (name, "default_value") << " ()"
                 << "{"
                 << "return " << unclash (name, "default_value") << "_;"
                 << "}";
            }
          }

          member_function_.traverse (m);
        }

      private:
        MemberFunction member_function_;
      };


      struct Complex : Traversal::Complex, protected virtual Context
      {
        Complex (Context& c, String const& name_ = L"")
            : Context (c),
              name (name_),
              member_ (c)
        {
          names_ >> member_;
        }

        virtual Void
        traverse (Type& c)
        {
          if (c.named ())
          {
            name = ename (c);

            // If renamed name is empty then we do not need to generate
            // anything for this type.
            //
            if (renamed_type (c, name) && !name)
              return;
          }

          enter_scope (name);
          String scoped_name (scope);
          leave_scope ();

          os << "// " << scoped_name << endl
             << "// " << endl
             << endl;


          // Go after anonymous types and generate accessors and modifiers
          // first.
          //
          {
            enter_scope (name);
            names (c, names_);
            leave_scope ();
          }

          os << endl;
        }

      private:
        String name;

        Traversal::Names names_;
        Member<Complex> member_;
      };


      // Goes after anonymous types via global element. Note that we don't
      // need to go after attributes because they can only be referenced
      // and in that case everything will be generated at the point of
      // reference.
      //
      struct GlobalMember : Traversal::Element,
                            GlobalElementBase,
                            protected virtual Context
      {
        GlobalMember (Context& c)
            : GlobalElementBase (c), Context (c)
        {
        }

        virtual Void
        traverse (Type& m)
        {
          if (!generate_p (m))
            return;

          SemanticGraph::Type& t (m.type ());

          if (!t.named ())
          {
            String const& name (ename (m));

            Traversal::Belongs belongs_;

            List list (*this, name);
            Union union_ (*this, name);
            Complex complex (*this, name);
            Enumeration enumeration (*this, name);

            belongs_ >> list;
            belongs_ >> union_;
            belongs_ >> complex;
            belongs_ >> enumeration;

            enter_scope (name);
            enter_scope (L"_xsd_" + name + L"_");

            t.context ().set ("seen", true);

            belongs (m, belongs_);

            t.context ().remove ("seen");

            leave_scope ();
            leave_scope ();
          }
        }
      };
    }

    Void
    generate_tree_inline (Context& ctx,
                          String const& expr,
                          UnsignedLong first,
                          UnsignedLong last)
    {
      // Generate includes.
      //
      if (ctx.options.value<CLI::generate_inline> ())
      {
        Traversal::Schema schema;
        Includes includes (ctx, expr);

        schema >> includes;

        schema.dispatch (ctx.schema_root);
      }


      Traversal::Schema schema;
      Traversal::Sources sources;
      Traversal::Names names_ns, names;
      Namespace ns (ctx, false, first, last);

      List list (ctx);
      Union union_ (ctx);
      Complex complex (ctx);
      Enumeration enumeration (ctx);
      GlobalMember member (ctx);

      schema >> sources >> schema;
      schema >> names_ns >> ns >> names;

      names >> list;
      names >> union_;
      names >> complex;
      names >> enumeration;
      names >> member;

      schema.dispatch (ctx.schema_root);
    }
  }
}
